/*
    This is an exapmle of an A* Pathfinding Algorithm written in Java/Greenfoot for my educational purposes and to prototype for future useage in games
    Copyright (C) 2011 Shaun Steenkamp

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>
 */
import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.*;
import java.lang.*;
import java.awt.Color;

/**
 * Finds the shortest path from start to finish whilst avoiding obsticals in the world.
 * 
 * This is where most of the processing takes place. Each node that is explored by the
 * pathfinder takes one act, although it would be very easy to just loop it a couple of times.
 * 
 * @author Shaun S.
 * @version 0.1 Pre-Alpha 050911
 */
public class pathfinder extends Actor
{
    private boolean searching = false;
    private boolean searched = false; // stores if a search has taken place.
    private pathNode start;
    private pathNode finish;
    private List<nodeSearch> unexplored = new LinkedList<nodeSearch>(); // open list
    private List<nodeSearch> explored = new LinkedList<nodeSearch>(); // closed list
    private double sqrt2 = Math.sqrt(2);
    private aStarPathfindingPxGrid world;
    
    private static final Color exploredColor = new Color(184, 150, 0);
    private static final Color routeColor = new Color(58, 184, 0);
    
    private static final int actsPerCycle = 15;
    private int currCycleCount = actsPerCycle;
    
    public pathfinder() {
        getImage().clear();
    }
    
    public void addedToWorld(World world) {
        this.world = (aStarPathfindingPxGrid) world;
    }
    
    public void act() 
    {
        if(searching) {
            if(currCycleCount >= actsPerCycle) {
                currCycleCount = 0;
                findPath();
            }
            currCycleCount++;
        } else if(Greenfoot.isKeyDown("space") && !searched) {
            startSearch();
        }
    }
    
    public void startSearch(pathNode start, pathNode finish) {
        this.start = start;
        this.finish = finish;
        searching = true;
        unexplored.add(
            new nodeSearch(start, 0, Double.MAX_VALUE, null));        
    }
    
    public void startSearch() {
        if(world.start != null && world.finish != null) {
            startSearch(world.start, world.finish);
        }
    }
    
    /**
     * this and expandNodes do the grunt of the actual pathfinding.
     */
    private void findPath() {
        if(unexplored.size()==0) {
            //if there are no more nodes to explre then we couldn't find a path.
            finishedPath(null);
        } else {
            ListIterator<nodeSearch> explorer = unexplored.listIterator();
            int nodePoz = 0;
            double lowestScore = Double.POSITIVE_INFINITY;
            // to start with the best score is inf because it needs to be higher than
            // any possible node in the list.
            nodeSearch bestNode = null;
            // go through each unexplored node searching for the lowest.
            while(explorer.hasNext()) {
                int nextNodePoz = explorer.nextIndex();
                nodeSearch node = explorer.next();
                if(node.getFVal() < lowestScore) {
                    lowestScore = node.getFVal();
                    bestNode = node;
                    nodePoz = nextNodePoz;
                }
            }
            // remove it from the unexplored list and add it to the explored list.
            unexplored.remove(nodePoz);
            explored.add(bestNode);
            // search all it's neighbours
            List<traversalCost> nodesFound = bestNode.getNode().getPathEdges();
            bestNode.getNode().changeColor(exploredColor);
            expandNodes(bestNode, nodesFound);
        }
    }
    
    private void expandNodes(nodeSearch parent, List<traversalCost> nodesFound) {
        for(traversalCost newNode : nodesFound) {
            if(parent.getParent() != null) {
                if(parent.getParent().getNode() == newNode.getNode()) {
                    continue; // returning to parent node is pointless!
                }
            }
            /*if(explored.indexOf(newNode) > -1) {
                continue;
            }*/
            double cost = parent.getCost() + newNode.getCost();
            boolean cheaperRouteExists = false;
            ListIterator<nodeSearch> explorer = unexplored.listIterator();
            // go through each node we have already found to see if this
            // is the same node
            while(explorer.hasNext()) {
                int nextNodePoz = explorer.nextIndex();
                nodeSearch node = explorer.next();
                // if we find a shorter route to the same node.
                if(node.getNode() == newNode.getNode()) {
                    if(cost < node.getCost()) {
                        // if our route to the node is faster then remove the other node
                        cheaperRouteExists = false;
                        unexplored.remove(nextNodePoz);
                    } else { // keep the other node
                        cheaperRouteExists = true;
                    }
                    break;
                }
            }
            ListIterator<nodeSearch> explorer2 = explored.listIterator();
            while(explorer2.hasNext()) {
                int nextNodePoz = explorer2.nextIndex();
                nodeSearch node = explorer2.next();
                if(node.getNode() == newNode.getNode()) {
                    if(cost < node.getCost()) {
                        cheaperRouteExists = false;
                        explored.remove(nextNodePoz);
                    } else {
                        cheaperRouteExists = true;
                    }
                    break;
                }
            }
            if(cheaperRouteExists) {
                //world.canvas.drawLine(parent.getNode(), newNode.getNode(), new Color(191, 0, 0));
                world.canvas.drawArrows(parent.getNode(), newNode.getNode(), new Color(191, 0, 0));
                continue;
            }
            // calculate the heuristic:
            double h = heuristicEuclidean(newNode.getNode(), finish);
            // calculate the F score
            double FVal = cost + h;
            // create a new nodeSearch object for this newly explored node.
            nodeSearch newNodeSearch = new nodeSearch(newNode.getNode(), cost, FVal, parent);
            world.canvas.drawLine(parent.getNode(), newNode.getNode(), exploredColor);
            world.canvas.drawArrows(parent.getNode(), newNode.getNode(), exploredColor);
            if(newNode.getNode() == finish) {
                // see if it is the finish node, if so we are done.
                finishedPath(newNodeSearch);
                System.out.println("finished: " + cost);
                searching = false;
            } else {
                unexplored.add(newNodeSearch);
            }
        }
    }
    
    /**
     * straight line tiebraker. Prefers nodes that are closest to direct line (as the crow flys)
     * between the start and the finish. (can have strange results with 4 directional movement and
     * obsticles, but still fastest path.)
     */
    private double tiebreaker(pathNode a, pathNode b) {
        double dx1 = a.getX() - b.getX();
        double dy1 = a.getY() - b.getY();
        double dx2 = start.getX() - b.getX();
        double dy2 = start.getY() - b.getY();
        return Math.abs(dx1*dy2 - dx2*dy1)*0.001;
        // I don't expect the path to ever be more than 1000 long so this will
        // never be big enough to bias the pathfinder to a non-optimal route.
    }
    
    /**
     * euclidean heuristic - estimates straight line distance from start to finish
     * good for any directional movement.
     */
    private double heuristicEuclidean(pathNode a, pathNode b) {
        return Math.sqrt( (b.getXcoord() - a.getXcoord())*(b.getXcoord() - a.getXcoord())
            + (b.getYcoord() - a.getYcoord())*(b.getYcoord() - a.getYcoord()));
    }
    
    /**
     * what to do when we have found a path........
     */
    private void finishedPath(nodeSearch endNode) {
        if(endNode == null) {
            System.out.println("No path found");
            searching = false;
            searched = true;
        } else {
            while(endNode.getParent() != null) {
                world.canvas.drawLine(endNode.getParent().getNode(), endNode.getNode(), routeColor);
                world.canvas.drawArrows(endNode.getParent().getNode(), endNode.getNode(), routeColor);
                endNode = endNode.getParent();
                endNode.getNode().changeColor(routeColor);
                Greenfoot.stop();
            }
        }
    }
}